@@ -9,8 +9,9 @@
 
 **********************************************************************/
 
-#include "internal.h"
+#include "ruby/encoding.h"
 #include "ruby/st.h"
+#include "internal.h"
 #include "symbol.h"
 #include "gc.h"
 #include "probes.h"
@@ -28,7 +29,7 @@ static ID register_static_symid_str(ID, VALUE);
 #define REGISTER_SYMID(id, name) register_static_symid((id), (name), strlen(name), enc)
 #include "id.c"
 
-#define is_identchar(p,e,enc) (rb_enc_isalnum((unsigned char)(*(p)),(enc)) || (*(p)) == '_' || !ISASCII(*(p)))
+#define is_identchar(p,e,enc) (ISALNUM((unsigned char)*(p)) || (*(p)) == '_' || !ISASCII(*(p)))
 
 #define op_tbl_count numberof(op_tbl)
 STATIC_ASSERT(op_tbl_name_size, sizeof(op_tbl[0].name) == 3);
@@ -51,7 +52,7 @@ Init_op_tbl(void)
     }
 }
 
-enum {ID_ENTRY_UNIT = 512};
+static const int ID_ENTRY_UNIT = 512;
 
 enum id_entry_type {
     ID_ENTRY_STR,
@@ -177,11 +178,11 @@ is_special_global_name(const char *m, const char *e, rb_encoding *enc)
 	}
     }
     else {
-	if (!rb_enc_isdigit(*m, enc)) return 0;
+	if (!ISDIGIT(*m)) return 0;
 	do {
 	    if (!ISASCII(*m)) mb = 1;
 	    ++m;
-	} while (m < e && rb_enc_isdigit(*m, enc));
+	} while (m < e && ISDIGIT(*m));
     }
     return m == e ? mb + 1 : 0;
 }
@@ -278,9 +279,9 @@ rb_enc_symname_type(const char *name, long len, rb_encoding *enc, unsigned int a
 	break;
 
       default:
-	type = rb_enc_isupper(*m, enc) ? ID_CONST : ID_LOCAL;
+	type = ISUPPER(*m) ? ID_CONST : ID_LOCAL;
       id:
-	if (m >= e || (*m != '_' && !rb_enc_isalpha(*m, enc) && ISASCII(*m))) {
+	if (m >= e || (*m != '_' && !ISALPHA(*m) && ISASCII(*m))) {
 	    if (len > 1 && *(e-1) == '=') {
 		type = rb_enc_symname_type(name, len-1, enc, allowed_attrset);
 		if (type != ID_ATTRSET) return ID_ATTRSET;
@@ -430,7 +431,8 @@ sym_check_asciionly(VALUE str)
     if (!rb_enc_asciicompat(rb_enc_get(str))) return FALSE;
     switch (rb_enc_str_coderange(str)) {
       case ENC_CODERANGE_BROKEN:
-	rb_raise(rb_eEncodingError, "invalid encoding symbol");
+	rb_raise(rb_eEncodingError, "invalid symbol in encoding %s :%+"PRIsVALUE,
+		 rb_enc_name(rb_enc_get(str)), str);
       case ENC_CODERANGE_7BIT:
 	return TRUE;
     }
@@ -674,13 +676,14 @@ rb_str_intern(VALUE str)
 #if USE_SYMBOL_GC
     enc = rb_enc_get(str);
     ascii = rb_usascii_encoding();
-    if (enc != ascii) {
-	if (sym_check_asciionly(str)) {
-	    str = rb_str_dup(str);
-	    rb_enc_associate(str, ascii);
-	    OBJ_FREEZE(str);
-	    enc = ascii;
-	}
+    if (enc != ascii && sym_check_asciionly(str)) {
+	str = rb_str_dup(str);
+	rb_enc_associate(str, ascii);
+	OBJ_FREEZE(str);
+	enc = ascii;
+    }
+    else {
+	str = rb_str_new_frozen(str);
     }
     str = rb_fstring(str);
     type = rb_str_symname_type(str, IDSET_ATTRSET_FOR_INTERN);
@@ -742,15 +745,7 @@ rb_sym2str(VALUE sym)
 VALUE
 rb_id2str(ID id)
 {
-    VALUE str;
-
-    if ((str = lookup_id_str(id)) != 0) {
-        if (RBASIC(str)->klass == 0)
-            RBASIC_SET_CLASS_RAW(str, rb_cString);
-	return str;
-    }
-
-    return 0;
+    return lookup_id_str(id);
 }
 
 const char *
@@ -1013,14 +1008,49 @@ rb_check_symbol_cstr(const char *ptr, long len, rb_encoding *enc)
     return Qnil;
 }
 
+#undef rb_sym_intern_cstr
+#undef rb_sym_intern_ascii_cstr
+#ifdef __clang__
+NOINLINE(VALUE rb_sym_intern(const char *ptr, long len, rb_encoding *enc));
+#else
+FUNC_MINIMIZED(VALUE rb_sym_intern(const char *ptr, long len, rb_encoding *enc));
+FUNC_MINIMIZED(VALUE rb_sym_intern_cstr(const char *ptr, rb_encoding *enc));
+FUNC_MINIMIZED(VALUE rb_sym_intern_ascii(const char *ptr, long len));
+FUNC_MINIMIZED(VALUE rb_sym_intern_ascii_cstr(const char *ptr));
+#endif
+
 VALUE
-rb_cstr_intern(const char *ptr, long len, rb_encoding *enc)
+rb_sym_intern(const char *ptr, long len, rb_encoding *enc)
 {
     struct RString fake_str;
     const VALUE name = rb_setup_fake_str(&fake_str, ptr, len, enc);
     return rb_str_intern(name);
 }
 
+VALUE
+rb_sym_intern_cstr(const char *ptr, rb_encoding *enc)
+{
+    return rb_sym_intern(ptr, strlen(ptr), enc);
+}
+
+VALUE
+rb_sym_intern_ascii(const char *ptr, long len)
+{
+    return rb_sym_intern(ptr, len, rb_usascii_encoding());
+}
+
+VALUE
+rb_sym_intern_ascii_cstr(const char *ptr)
+{
+    return rb_sym_intern_ascii(ptr, strlen(ptr));
+}
+
+VALUE
+rb_to_symbol_type(VALUE obj)
+{
+    return rb_convert_type_with_id(obj, T_SYMBOL, "Symbol", idTo_sym);
+}
+
 static ID
 attrsetname_to_attr_id(VALUE name)
 {
